#!/usr/bin/env python

from osgeo import gdal,ogr,osr
import os
import numpy as np
import math
import sys
import csv
gdal.UseExceptions()  # Enable exception support
def main():
    if len(sys.argv) == 1:
        print('ERROR: meshstats requires one argument, the directory of the output from a mesher run (e.g.,  mesher.py Bow)')
        return

    # Get name of configuration file/module
    mesher_output_dir = sys.argv[-1]

    print('Reading in files')

    #############

    base_name =os.path.basename(os.path.normpath(mesher_output_dir))
    raster_file = os.path.normpath(mesher_output_dir)+'/'+base_name+'_projected.tif'
    shp_file = os.path.normpath(mesher_output_dir)+'/'+base_name+'_USM.shp'

    base_shp_name =os.path.basename(os.path.normpath(shp_file))

    raster_ds = gdal.Open(raster_file)
    if raster_ds is None:
        print('Could not open %s' % raster_ds)
        exit(1)


    rb = raster_ds.GetRasterBand(1)
    src_array = rb.ReadAsArray(0,0,raster_ds.RasterXSize-1,raster_ds.RasterYSize-1)
    masked = np.ma.masked_where(src_array == rb.GetNoDataValue(),src_array )
    c = masked.count()

    src_array = None
    masked = None
    rb = None
    if c == 0:
        print ("Only no data present in raster")
        exit(1)


    print ("Number of raster cells = " + str(c))

    mesh = ogr.Open(shp_file, update=True)
    if mesh is None:
        print( 'Could not open %s' % mesh)
        exit(1)

    layer = mesh.GetLayer()
    num_elem = layer.GetFeatureCount()
    print( "Number of triangles = %d" % num_elem)

    print ("# triangles v. raster = " + str( round(num_elem / float(c) * 100.,2)) + '%')

    layerDefinition = layer.GetLayerDefn()
    field_names = []
    for i in range(layerDefinition.GetFieldCount()):
        field_names.append(layerDefinition.GetFieldDefn(i).GetName())

    if 'area' not in field_names:
        layer.CreateField(ogr.FieldDefn('area', ogr.OFTReal))
    if 'min_angle' not in field_names:
        layer.CreateField(ogr.FieldDefn('min_angle', ogr.OFTReal))
    if 'max_angle' not in field_names:
        layer.CreateField(ogr.FieldDefn('max_angle', ogr.OFTReal))
    if 'error' not in field_names:
        layer.CreateField(ogr.FieldDefn('error', ogr.OFTReal))

    i=0
    with open(base_shp_name+'_stats.csv','w') as f:
        header = ["error", "area", "min_angle","max_angle"]
        writer = csv.writer(f)
        writer.writerow(header)

        for feature in layer:
            printProgress(i,num_elem)

            geom = feature.GetGeometryRef()
            ring = geom.GetGeometryRef(0)

            # tri_id = feature.GetField('triangle')

            geom_area = geom.GetArea()
            feature.SetField('area', geom_area)

            angle = tri_angles(raster_ds, ring)
            #if angles comes back none, we've hit an edge case where the triangle slightly sits outside of the domain.
            #this would have been fixed in the output mesh from mesher, but just ignore it here

            if angle is not None:
              feature.SetField('min_angle',min(angle))
              feature.SetField('max_angle',max(angle))

            rmse = tri_rmse(raster_ds, feature)
            if rmse is not None:
                feature.SetField('error',rmse)

            layer.SetFeature(feature)

            row = [rmse, geom_area, min(angle), max(angle)]
            writer.writerow(row)

            i+=1

    mesh= None




def tri_angles(raster_ds, ring):
    x1, y1, z = ring.GetPoint(0)  # this z is 0,
    z1 = extract_point(raster_ds, x1, y1)
    x2, y2, z = ring.GetPoint(1)
    z2 = extract_point(raster_ds, x2, y2)
    x3, y3, z = ring.GetPoint(2)
    z3 = extract_point(raster_ds, x3, y3)

    rb = raster_ds.GetRasterBand(1)
    if z1 == rb.GetNoDataValue() or z2 == rb.GetNoDataValue() or z3 == rb.GetNoDataValue():
        return None

    # vector length
    def vl(v):
        return math.sqrt(v[0] ** 2 + v[1] ** 2)

    # dot product
    def dp(u, v):
        return u[0] * v[0] + u[1] * v[1]

    u = (y3 - y1, x3 - x1)
    v = (y2 - y1, x2 - x1)
    angle32 = math.acos(dp(u, v) / (vl(u) * vl(v))) * 180. / math.pi
    u = (y1 - y3, x1 - x3)
    v = (y2 - y3, x2 - x3)
    angle12 = math.acos(dp(u, v) / (vl(u) * vl(v))) * 180. / math.pi
    u = (y1 - y2, x1 - x2)
    v = (y3 - y2, x3 - x2)
    angle13 = math.acos(dp(u, v) / (vl(u) * vl(v))) * 180. / math.pi

    return [angle32, angle12, angle13]


def tri_rmse(raster_ds,feature):
    geom = feature.GetGeometryRef()
    ring = geom.GetGeometryRef(0)

    new_gt, rtri,xsize,ysize = rasterize_elem(raster_ds, feature)

    #get triangle verticies
    x1, y1, z = ring.GetPoint(0)  # this z is 0,
    z1 = extract_point(raster_ds, x1, y1)
    x1, y1 = xyToPixel(new_gt,x1,y1,xsize,ysize)

    x2, y2, z = ring.GetPoint(1)
    z2 = extract_point(raster_ds, x2, y2)
    x2, y2 = xyToPixel(new_gt,x2,y2,xsize,ysize)

    x3, y3, z = ring.GetPoint(2)
    z3 = extract_point(raster_ds, x3, y3)
    x3, y3 = xyToPixel(new_gt,x3,y3,xsize,ysize)

    rb = raster_ds.GetRasterBand(1)
    if z1 == rb.GetNoDataValue() or z2 == rb.GetNoDataValue() or z3 == rb.GetNoDataValue():
        return None

    u1 = x2 - x1
    u2 = y2 - y1
    u3 = z2 - z1

    v1 = x3 - x1
    v2 = y3 - y1
    v3 = z3 - z1

    a = u2 * v3 - v2 * u3
    b = v1 * u3 - u1 * v3
    c = u1 * v2 - v1 * u2

    d =a*x1 + b*y1 + c*z1

    rmse = 0
    n = 0
    if c == 0:    #collinear
        return None

    max_diff = -9999
    for y in range(rtri.shape[0]):
        for x in range(rtri.shape[1]):
            if not np.ma.is_masked(rtri[y,x]):
                pred = -(a*x+b*y-d)/c #plane eqn solved for z. allows us to predict z values via x,y coords
                value = rtri[y,x]
                diff = value - pred
                rmse += diff ** 2
                n+=1
                if abs(diff) > max_diff:
                    max_diff = abs(diff)

    if n==0:
        return None
    rmse /= n
    rmse = math.sqrt(rmse)

    # return max_diff
    # rmse = ((z1+z2+z3)/3)-float(rtri.mean())
    return abs(rmse)

#Zonal stats from here https://gist.github.com/perrygeo/5667173
def bbox_to_pixel_offsets(gt, bbox, rasterXsize, rasterYsize):
    originX = gt[0]
    originY = gt[3]
    pixel_width = gt[1]
    pixel_height = gt[5]
    x1 = int((bbox[0] - originX) / pixel_width)
    x2 = int((bbox[1] - originX) / pixel_width) + 1

    y1 = int((bbox[3] - originY) / pixel_height)
    y2 = int((bbox[2] - originY) / pixel_height) + 1

    xsize = x2 - x1
    ysize = y2 - y1

    #only apply this correction if we are touching the underlying raster.
    # if x1 < rasterXsize and y1 < rasterYsize:
        #deal with small out of bounds
    if x1 < 0:
        x1 = 0

    if y1 < 0:
        y1 = 0

    if x1 + xsize >= rasterXsize:
        xsize = rasterXsize-x1

    if y1 + ysize >= rasterYsize:
        ysize = rasterYsize-y1

    return x1, y1, xsize, ysize

def xyToPixel( gt, x,  y,  max_x,  max_y):
    # adfGeoTransform[0] /* top left x */
    # adfGeoTransform[1] /* w-e pixel resolution */
    # adfGeoTransform[2] /* 0 */
    # adfGeoTransform[3] /* top left y */
    # adfGeoTransform[4] /* 0 */
    # adfGeoTransform[5] /* n-s pixel resolution (negative value) */

    px = int((x - gt[0]) / gt[1])  # x pixel
    py = int((y - gt[3]) / gt[5])  # y pixel

    #out of bound issue when we are off by one
    if px >= max_x:
        px = max_x - 1

    if py >= max_y:
        py = max_y - 1

    if py < 0:
        py = 0

    if px <0 :
        px = 0

    return px, py

def rasterize_elem(raster, feature):

    wkt = raster.GetProjection()
    srs = osr.SpatialReference()
    srs.ImportFromWkt(wkt)
    gt = raster.GetGeoTransform()
    rb = raster.GetRasterBand(1)
    src_offset = bbox_to_pixel_offsets(gt, feature.geometry().GetEnvelope(), raster.RasterXSize, raster.RasterYSize)

    src_array = rb.ReadAsArray(*src_offset)
    # calculate new geotransform of the feature subset
    new_gt = (
        (gt[0] + (src_offset[0] * gt[1])),
        gt[1],
        0.0,
        (gt[3] + (src_offset[1] * gt[5])),
        0.0,
        gt[5]
    )

    # Create a temporary vector layer in memory
    mem_drv = ogr.GetDriverByName('Memory')
    mem_ds = mem_drv.CreateDataSource('out')
    # mem_drv = ogr.GetDriverByName('ESRI Shapefile')
    # mem_ds = mem_drv.CreateDataSource('rastertri.shp')

    mem_layer = mem_ds.CreateLayer('poly', srs, ogr.wkbPolygon)
    mem_layer.CreateFeature(feature.Clone())

    # Rasterize it
    driver = gdal.GetDriverByName('MEM')
    rvds = driver.Create('', src_offset[2], src_offset[3], 1, gdal.GDT_Byte)
    # driver = gdal.GetDriverByName('GTiff')
    # rvds = driver.Create('rasterized.tif', src_offset[2], src_offset[3], 1, gdal.GDT_Byte)
    rvds.SetGeoTransform(new_gt)

    gdal.RasterizeLayer(rvds, [1], mem_layer, burn_values=[1], options=['ALL_TOUCHED=TRUE'])
    rv_array = rvds.ReadAsArray()  # holds a mask of where the triangle is on the raster

    # Mask the source data array with our current feature
    # we take the logical_not to flip 0<->1 to get the correct mask effect
    # we also mask out nodata values explictly
    masked = np.ma.MaskedArray(
        src_array,
        mask=np.logical_or(
            src_array == rb.GetNoDataValue(),
            np.logical_not(rv_array)
        )
    )

    return new_gt,masked,src_offset[2],src_offset[3]
def extract_point(raster,mx,my):
    # Convert from map to pixel coordinates.
    # Only works for geotransforms with no rotation.

    rb = raster.GetRasterBand(1)
    gt = raster.GetGeoTransform()

    px = int((mx - gt[0]) / gt[1])  # x pixel
    py = int((my - gt[3]) / gt[5])  # y pixel

    #boundary verticies from Triangle can end up outside of the domain by 1 pixel.
    #if we adjusted back by  1 pixel to get the dz value, it's no problem and still gives a good boundary
    if px == raster.RasterXSize:
        px = px -1
    if py == raster.RasterYSize:
        py = py - 1
    mz = rb.ReadAsArray(px, py, 1, 1)
    mz = float(mz.flatten()[0])

    if mz == rb.GetNoDataValue():
        dx = 1
        dy = 1

        #attempt to pick points from the surrounding 8
        z1 = rb.ReadAsArray(px - dx if px > 0 else px, py, 1, 1)
        z1 = z1[0] if z1 is not None else rb.GetNoDataValue()

        z2 = rb.ReadAsArray(px + dx if px < raster.RasterXSize else px, py, 1, 1)
        z2 = z2[0] if z2 is not None else rb.GetNoDataValue()

        z3 = rb.ReadAsArray(px, py + dy if py < raster.RasterYSize else py, 1, 1)
        z3 = z3[0] if z3 is not None else rb.GetNoDataValue()

        z4 = rb.ReadAsArray(px, py - dy if py > 0 else py, 1, 1)[0]
        z4 = z4[0] if z4 is not None else rb.GetNoDataValue()

        z5 = rb.ReadAsArray(px - dx if px > 0 else px,                  py - dy if py > 0 else py, 1, 1)
        z5 = z5[0] if z5 is not None else rb.GetNoDataValue()

        z6 = rb.ReadAsArray(px + dx if px < raster.RasterXSize else px, py - dy if py > 0 else py, 1, 1)
        z6 = z6[0] if z6 is not None else rb.GetNoDataValue()

        z7 = rb.ReadAsArray(px - dx if px > 0 else px,                  py + dy if py < raster.RasterYSize else py, 1, 1)
        z7 = z7[0] if z7 is not None else rb.GetNoDataValue()

        z8 = rb.ReadAsArray(px + dx if px < raster.RasterXSize else px, py - dy if py > 0 else py, 1, 1)
        z8 = z8[0] if z8 is not None else rb.GetNoDataValue()

        z=[z1, z2, z3, z4,z5,z6,z7,z8]
        z = [x for x in z if x != rb.GetNoDataValue()]

        if len(z) == 0:
            print ('Warning: The point (%f,%f) and its 8-neighbours lies outside of the DEM domain' %(mx,my))
            return rb.GetNoDataValue()
            #exit(1)

        mz = float(np.mean(z))
    return mz


# Print iterations progress
# http://stackoverflow.com/a/34325723/410074
def printProgress(iteration, total, prefix='', suffix='', decimals=2, barLength=100):
    """
    Call in a loop to create terminal progress bar
    @params:
        iterations  - Required  : current iteration (Int)
        total       - Required  : total iterations (Int)
        prefix      - Optional  : prefix string (Str)
        suffix      - Optional  : suffix string (Str)
    """
    filledLength = int(round(barLength * iteration / float(total)))
    percents = round(100.00 * (iteration / float(total)), decimals)
    bar = '#' * filledLength + '-' * (barLength - filledLength)
    sys.stdout.write('%s [%s] %s%s %s\r' % (prefix, bar, percents, '%', suffix)),
    sys.stdout.flush()
    if iteration == total:
        print("\n")

if __name__ == "__main__":

    main()



#!/usr/bin/env python

import json
import vtk

import sys

if len(sys.argv) != 4:
    print('requires 3 arguments: .mesh file, .param file, output vtu file')
    exit(-1)

#open topology file
with open(sys.argv[1]) as f:
    mesh = json.load(f)

#open parameter file
with open(sys.argv[2]) as f:
    param = json.load(f)

#init the vtk datastructure
vtu = vtk.vtkUnstructuredGrid()
vtuwriter = vtk.vtkXMLUnstructuredGridWriter()
vtuwriter.SetFileName(sys.argv[3])
vtuwriter.SetInputData(vtu)

vtu_points = vtk.vtkPoints()
vtu_triangles = vtk.vtkCellArray()

vtu_points.SetNumberOfPoints(len(mesh['mesh']['vertex']))

# elevation isn't stored in the param or mesh file, so compute from the z like mesher does
vtu_cells  = {'Elevation': vtk.vtkFloatArray()}
vtu_cells['Elevation'].SetName('elevation')

# loop over the mesh elements and reconstruct the triangle
for e in mesh['mesh']['elem']:
    v0 = e[0]
    v1 = e[1]
    v2 = e[2]

    vtu_points.SetPoint(v0, mesh['mesh']['vertex'][v0][0], mesh['mesh']['vertex'][v0][1], mesh['mesh']['vertex'][v0][2])
    vtu_points.SetPoint(v1, mesh['mesh']['vertex'][v1][0], mesh['mesh']['vertex'][v1][1], mesh['mesh']['vertex'][v1][2])
    vtu_points.SetPoint(v2, mesh['mesh']['vertex'][v2][0], mesh['mesh']['vertex'][v2][1], mesh['mesh']['vertex'][v2][2])

    triangle = vtk.vtkTriangle()
    triangle.GetPointIds().SetId(0, v0)
    triangle.GetPointIds().SetId(1, v1)
    triangle.GetPointIds().SetId(2, v2)

    vtu_triangles.InsertNextCell(triangle)

    vtu_cells['Elevation'].InsertNextTuple1((mesh['mesh']['vertex'][v0][2] +
                                                             mesh['mesh']['vertex'][v1][2] +
                                                             mesh['mesh']['vertex'][v2][2]) / 3.)

# insert all the parameters from the param file
for key, data in param.items():
    k =  key

    #sanity check to ensure we don't add a 0 length
    if len(param[k]) == 0 :
        continue
    vtu_cells[k]=vtk.vtkFloatArray()
    vtu_cells[k].SetName(k)

    for e in param[k]:
        vtu_cells[k].InsertNextTuple1(e)

# write the proj4 wkt output
proj4 = vtk.vtkStringArray()
proj4.SetNumberOfComponents(1);
proj4.SetName("proj4");
proj4.InsertNextValue(mesh['mesh']['proj4']);

vtu.GetFieldData().AddArray(proj4);

vtu.SetPoints(vtu_points)
vtu.SetCells(vtk.VTK_TRIANGLE,vtu_triangles)
for p in vtu_cells.values():
    vtu.GetCellData().AddArray(p)
vtuwriter.Write()





#!/usr/bin/env python
import json
from pathlib import Path

# Set up CL arguments
import argparse
parser = argparse.ArgumentParser()

parser.add_argument("-o", "--outfile", required=False,
                help="File for output. Default is an filename from all input files.")

parser.add_argument("infile",
                    nargs='+',
                    help="Files containing parameters to be merged.")


def append_new_param_to_file(args):
    """Read dictionary of arguments, find parameters, add new parameters to reference json file"""

    params = args["infile"]

    # load the first to append to
    with open(params[0]) as f:
        param_merged = json.load( f )

    for i in range(1, len(params)):
        with open(params[i]) as f:
            param_in = json.load(f)
            # Add parameters to param_ref
            param_merged.update(param_in)

    lengths = []

    print('Merged param file has:')
    for param, data in param_merged.items():

        l = len(data)
        lengths.append(l)
        print(F"{param}: {l}")

    if len(set(lengths)) > 1:
        print('ERROR: Not all parameters have the same length, which they must.')

    with open(args["outfile"],'w') as f:
        json.dump(param_merged, f, indent=4)


if __name__=="__main__":

    # Parse the input arguments
    args = vars(parser.parse_args())
    if args["outfile"] is None:
        name = ''
        for f in args["infile"]:
            filename = Path(f).resolve().stem
            name = name + filename + '+'
        args["outfile"] = '%s.param' % name[:-1]

    append_new_param_to_file(args)



#!/usr/bin/env python
# Mesher
# Copyright (C) 2017 Christopher Marsh

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from osgeo import gdal, ogr, osr
import subprocess
import re
import json
import os
import numpy as np
from functools import partial
import sys
import shutil
import importlib
import vtk
import warnings
import uuid
import cloudpickle
from concurrent import futures
import time

# This reverts to Python <= 3.7 behaviours. It "worked" before...
# Long term it's almost certainly a Bad Idea and should be resolved.
# > Changed in version 3.8: On macOS, the spawn start method is now the default.
# >The fork start method should be considered unsafe as it can lead to crashes of the subprocess. See bpo-33725.
# Setting this to 'spawn' results in the imported config file not being available for the child processes
# Ref https://bugs.python.org/issue33725
import multiprocessing as mp
mp.set_start_method("fork")


gdal.UseExceptions()  # Enable exception support
ogr.UseExceptions()  # Enable exception support
osr.UseExceptions()  # Enable exception support


def main():
    # load user configurable parameters here
    # Check user defined configuration file

    if len(sys.argv) == 1:
        print('ERROR: mesher.py requires one argument [configuration file] (i.e. mesher.py Bow.py)')
        return

    # Get name of configuration file/module
    configfile = sys.argv[-1]

    # Load in configuration file as module
    X = importlib.machinery.SourceFileLoader('config', configfile)
    X = X.load_module()

    # get config file dire
    cwd = os.path.join(os.getcwd(), os.path.dirname(configfile))
    os.chdir(cwd)

    dem_filename = X.dem_filename.strip()
    max_area = X.max_area

    # load any user given parameter files
    parameter_files = {}
    if hasattr(X, 'parameter_files'):
        parameter_files = X.parameter_files

        for key, data in parameter_files.items():
            if isinstance(data['file'], list):
                for i in range(len(data['file'])):
                    parameter_files[key]['file'][i] = parameter_files[key]['file'][i].strip()

    # initial conditions to apply to the triangles with
    initial_conditions = {}
    if hasattr(X, 'initial_conditions'):
        initial_conditions = X.initial_conditions

        for key, data in initial_conditions.items():
            initial_conditions[key]['file'] = initial_conditions[key]['file'].strip()

    # any extra user-specific constraints. E.g., streams
    constraints = {}
    if hasattr(X, 'constraints'):
        constraints = X.constraints

        for key, data in constraints.items():
            constraints[key]['file'] = constraints[key]['file'].strip()

    # simplify applies a simplification routine in GDAL to the outer boundary such
    # that the error between the simplified geom and the original is no more than simplify_tol.
    # This allows the mesh generator more lee way in not creating small triangles allong the boundary
    simplify = False
    simplify_tol = 10
    if hasattr(X, 'simplify'):
        simplify = X.simplify

        if hasattr(X, 'simplify_tol'):
            simplify_tol = X.simplify_tol

    # simplify buffer contracts the outer geometry by bufferDist to help avoid creating small triangles
    bufferDist = -10  # default to -10, will only trigger is simplify is set to T
    if hasattr(X, 'simplify_buffer'):
        bufferDist = X.simplify_buffer

    # Can set simplify_buffer to 0, but we can also just disable this here
    no_simplify_buffer = False
    if hasattr(X, 'no_simplify_buffer'):
        bufferDist = X.no_simplify_buffer

    if bufferDist > 0:
        print('Buffer must be < 0 as we need to shrink the extent.')
        exit(-1)

    # Enable lloyd iterations "The goal of this mesh optimization is to improve the angles inside the mesh,
    # and make them as close as possible to 60 degrees." 100 iterations is a suggested amount
    # https://doc.cgal.org/latest/Mesh_2/index.html#secMesh_2_optimization
    lloyd_itr = 0
    if hasattr(X, 'lloyd_itr'):
        lloyd_itr = X.lloyd_itr

    # maximum tolerance, as measured by the error metric, between the triangle and the elevation raster.
    # -1 skips the tolerance check -- useful for producing uniform triangles
    max_tolerance = None
    if hasattr(X, 'max_tolerance'):
        max_tolerance = X.max_tolerance

    # Choice of error metric.
    # Can be RMSE or tolerance
    errormetric = 'rmse'
    if hasattr(X, 'errormetric'):
        errormetric = X.errormetric

    # if a mesh was already generated, and only applying a new parametrization is required,
    # enabling this skips the mesh generation step
    reuse_mesh = False
    if hasattr(X, 'reuse_mesh'):
        reuse_mesh = X.reuse_mesh

    mesher_path = os.path.dirname(os.path.abspath(__file__)) + '/mesher'

    # uses GDAL to fill holes
    fill_holes = False
    if hasattr(X, 'fill_holes'):
        fill_holes = X.fill_holes


    # look for MESHER_EXE as an environment variable. Defining the mesher path in the config file takes precedenc
    # over this
    using_mesher_environ = False
    try:
        mesher_path = os.environ['MESHER_EXE']
        using_mesher_environ = True
    except KeyError as E:
        pass

    # path to mesher executable
    if hasattr(X, 'mesher_path'):
        mesher_path = X.mesher_path

        if using_mesher_environ:
            warnings.warn(
                "Warning: mesher binary path defined in env var and in configuration file. Using the mesher path from "
                "the configuration file")

            # enable verbose output for debugging
    verbose = False
    if hasattr(X, 'verbose'):
        verbose = X.verbose

    user_output_dir = cwd + os.path.sep

    # output to the specific directory, instead of the root dir of the calling python script
    if hasattr(X, 'user_output_dir'):
        user_output_dir += X.user_output_dir
    else:
        # use the config filename as output path
        (file, ext) = os.path.splitext(os.path.basename(configfile))
        user_output_dir += file

    if user_output_dir[-1] is not os.path.sep:
        user_output_dir += os.path.sep

    # should we write a shape file of the USM, pretty costly on the big meshes
    output_write_shp = True
    if hasattr(X, 'write_shp'):
        output_write_shp = X.write_shp

    output_write_vtu = True
    if hasattr(X, 'write_vtu'):
        output_write_vtu = X.write_vtu

    # Use the input file's projection.
    # This is useful for preserving a UTM input. Does not work if the input file is geographic.
    use_input_prj = True
    if hasattr(X, 'use_input_prj'):
        use_input_prj = X.use_input_prj

    # Do smoothing of the input DEM to create a more smooth mesh. This can help if the DEM quality is poor or if
    # triangles close to the elevation raster cell size is required
    do_smoothing = False
    if hasattr(X, 'do_smoothing'):
        do_smoothing = X.do_smoothing

    # Smoothing factor for above option.
    scaling_factor = 2.0
    if hasattr(X, 'smoothing_scaling_factor'):
        scaling_factor = X.smoothing_scaling_factor

    # number of iterations to smooth over, each smoothing using cubic spline,
    # and resamples by iter * scaling_factor for each iteration
    max_smooth_iter = 1
    if hasattr(X, 'max_smooth_iter'):
        max_smooth_iter = X.max_smooth_iter

    # use the convex combination of weights method
    user_no_weights = False
    weight_threshold = 0.5
    if hasattr(X, 'weight_threshold'):
        weight_threshold = X.weight_threshold

    if hasattr(X, 'use_weights'):
        if not X.use_weights:
            user_no_weights = True

    wkt_out = "PROJCS[\"North_America_Albers_Equal_Area_Conic\"," \
              "     GEOGCS[\"GCS_North_American_1983\"," \
              "         DATUM[\"North_American_Datum_1983\"," \
              "             SPHEROID[\"GRS_1980\",6378137,298.257222101]]," \
              "         PRIMEM[\"Greenwich\",0]," \
              "         UNIT[\"Degree\",0.017453292519943295]]," \
              "     PROJECTION[\"Albers_Conic_Equal_Area\"]," \
              "     PARAMETER[\"False_Easting\",0]," \
              "     PARAMETER[\"False_Northing\",0]," \
              "     PARAMETER[\"longitude_of_center\",-96]," \
              "     PARAMETER[\"Standard_Parallel_1\",20]," \
              "     PARAMETER[\"Standard_Parallel_2\",60]," \
              "     PARAMETER[\"latitude_of_center\",40]," \
              "     UNIT[\"Meter\",1]," \
              "     AUTHORITY[\"EPSG\",\"102008\"]]"
    if hasattr(X, 'wkt_out'):
        wkt_out = X.wkt_out

    # use a custom extent instead of the entire input DEM to define the meshing region
    # extent = [xmin, ymin, xmax, ymax] in source SRS

    extent = None
    if hasattr(X, 'extent'):
        extent = X.extent

    nworkers = os.cpu_count() or 1

    # on linux we can ensure that we respect cpu affinity
    if 'sched_getaffinity' in dir(os):
        nworkers = len(os.sched_getaffinity(0))

    if hasattr(X, 'nworkers'):
        nworkers = X.nworkers

    # GDAL workers are for the initial regularize call where multiple gdalwarp processes are started
    # these are memory heavy so this can be limited if required
    nworkers_gdal = nworkers
    if hasattr(X, 'nworkers_gdal'):
        nworkers_gdal = X.nworkers_gdal
    try:
        nworkers_gdal = os.environ['MESHER_NWORKERS_GDAL']

        if hasattr(X, 'nworkers_gdal'):
            print('Warning: Overridding configfile nworkers_gdal with environment variable MESHER_NWORKERS_GDAL')
    except KeyError as E:
        pass

    try:
        nworkers = os.environ['MESHER_NWORKERS']

        if hasattr(X, 'nworkers'):
            print('Warning: Overridding configfile nworkers with environment variable MESHER_NWORKERS')
    except KeyError as E:
        pass
    nworkers = int(nworkers)
    nworkers_gdal = int(nworkers_gdal)

    try:
        cache = os.environ['GDAL_CACHEMAX']
        print(
            'Warning: GDAL_CACHEMAX={0}. Ensure this is reasonable for the number of parallel processes specified'.format(
                cache))
    except:
        # the user doesn't have the GDAL_CACHEMAX defined, which is good. Set it to a reasonable value
        os.environ['GDAL_CACHEMAX'] = "{0}%".format(30.0 / nworkers)

    print('Using {0} CPUs and per-CPU memory cache = {1}'.format(nworkers, os.environ['GDAL_CACHEMAX']))
    ########################################################

    # we need to make sure we pickup the right paths to all the gdal scripts
    gdal_prefix = ''
    try:
        gdal_prefix = subprocess.run(["gdal-config", "--prefix"], stdout=subprocess.PIPE).stdout.decode()
        gdal_prefix = gdal_prefix.replace('\n', '')
        gdal_prefix += '/bin/'
    except:
        raise BaseException(""" ERROR: Could not find gdal-config, please ensure it is installed and on $PATH """)

    base_name = os.path.basename(dem_filename)
    base_name = os.path.splitext(base_name)[0]

    base_dir = user_output_dir + base_name + os.path.sep

    # Delete previous dir (if exists)
    if os.path.isdir(base_dir) and not reuse_mesh:
        shutil.rmtree(base_dir, ignore_errors=True)

    # these have to be separate ifs for the logic to work correctly
    if not reuse_mesh:
        # make new output dir
        os.makedirs(base_dir)

    # we want to reuse an already generated mesh, but we will need to clean up the shp file as gdal won't overwrite
    # an existing one
    if reuse_mesh:
        try:
            os.remove(base_dir + base_name + '_USM.shp')
        except OSError:
            pass  # If the folder doesn't exist, don't worry

    try:
        src_ds = gdal.Open(dem_filename)
    except RuntimeError as e:
        print('Unable to open file ' + dem_filename)
        raise e

    if src_ds.GetProjection() == '':
        raise RuntimeError("Input DEM must have spatial reference information.")

    # if we are using the input DEM's projection, grab it here from the source DEM
    if use_input_prj:
        wkt_out = src_ds.GetProjection()

    srs_out = osr.SpatialReference()

    srs_out.ImportFromWkt(wkt_out)

    if do_smoothing:
        output_file_name = '_projected_0.tif'
    else:
        output_file_name = '_projected.tif'

    ext_str = ''

    # clip to a user-specified extent
    if extent is not None:
        src_ds = gdal.Open(dem_filename)
        wkt = src_ds.GetProjection()
        srs = osr.SpatialReference()

        srs.ImportFromWkt(wkt)

        ext_str = ' -te %s %s %s %s -te_srs \"%s\" ' % (extent[0], extent[1], extent[2], extent[3], srs.ExportToProj4())
        src_ds = None

    e = 'GDAL_CACHEMAX=\"5%%\" %sgdalwarp %s %s -ot Float32 -multi -overwrite -dstnodata -9999 -t_srs \"%s\"' + ext_str
    subprocess.check_call([e % (gdal_prefix,
                                dem_filename, base_dir + base_name + output_file_name, srs_out.ExportToProj4())],
                          shell=True)

    if fill_holes:
        gt = src_ds.GetGeoTransform()

        pixel_width = gt[1]
        pixel_height = gt[5]

        dist = max([pixel_height, pixel_width]) * 5
        exec_str = '%sgdal_fillnodata.py %s -md %d' % (gdal_prefix, base_dir + base_name + output_file_name, dist)
        subprocess.check_call([exec_str], shell=True)

    src_ds = gdal.Open(base_dir + base_name + output_file_name)

    if src_ds is None:
        print('Unable to open %s' % dem_filename)
        exit(1)

    #obtain new values for this after all the projection, etc
    gt = src_ds.GetGeoTransform()

    pixel_width = gt[1]
    pixel_height = gt[5]

    if hasattr(X, 'min_area'):
        min_area = X.min_area
    else:
        # if the user doesn't specify, then limit to the underlying resolution. No point going past this!
        min_area = abs(
            pixel_width * pixel_height)

    if do_smoothing:
        for itr in range(max_smooth_iter):

            in_name = base_dir + base_name + '_projected_%d.tif' % itr
            out_name = base_dir + base_name + '_projected_%d.tif' % (itr + 1)

            if itr + 1 == max_smooth_iter:  # last iteration, change output
                out_name = base_dir + base_name + '_projected.tif'

            subprocess.check_call(
                [
                    'GDAL_CACHEMAX=\"5%%\" %sgdalwarp %s %s -ot Float32 -multi  -overwrite -dstnodata -9999 -r cubicspline -tr %s %s' % (
                        gdal_prefix,
                        in_name, out_name, abs(pixel_width) / scaling_factor,
                        abs(pixel_height) / scaling_factor)], shell=True)

            scaling_factor *= (itr + 1)

    # now, reopen the file
    src_ds = gdal.Open(base_dir + base_name + '_projected.tif')

    # create the spatial reference from the raster dataset
    wkt = src_ds.GetProjection()
    srs = osr.SpatialReference()

    srs.ImportFromWkt(wkt)

    is_geographic = srs.IsGeographic()

    if is_geographic and simplify:
        print('Error: Simplify=True and a geographic output CRS is not currently supported.')
        exit(1)

    gt = src_ds.GetGeoTransform()
    # x,y origin
    xmin = gt[0]
    ymax = gt[3]

    pixel_width = gt[1]
    pixel_height = gt[5]

    xmax = xmin + pixel_width * src_ds.RasterXSize
    ymin = ymax + pixel_height * src_ds.RasterYSize  # pixel_height is negative

    exec_str = f'%sgdalwarp %s %s -ot Float32 -overwrite -multi -dstnodata -9999 -t_srs "%s" -te %s %s %s %s  -r '

    total_weights_param, use_weights_param = regularize_inputs(base_dir, exec_str, gdal_prefix, parameter_files,
                                                               pixel_height, pixel_width,
                                                               srs_out, xmax, xmin, ymax, ymin, fill_holes, nworkers_gdal)

    total_weights_ic, use_weights_ic = regularize_inputs(base_dir, exec_str, gdal_prefix, initial_conditions,
                                                         pixel_height, pixel_width,
                                                         srs_out, xmax, xmin, ymax, ymin, fill_holes, nworkers_gdal)

    use_weights = use_weights_param or use_weights_ic

    topo_weight = 1 - (total_weights_param + total_weights_ic)

    # over ride with user settings
    if user_no_weights:
        use_weights = False

    if use_weights and topo_weight < 0:
        raise RuntimeError("Parameter weights must equal 1")

    plgs_shp = base_name + '.shp'

    dem = src_ds.GetRasterBand(1)

    # Create a mask raster that has a uniform value for all cells read all elevation data. Might be worth changing to
    # the for-loop approach we use below so we don't have to read in all into ram. some raster
    Z = dem.ReadAsArray()

    # set all the non-nodata values to our mask value
    Z[Z != dem.GetNoDataValue()] = 1

    # save to file
    (x, y) = Z.shape
    driver = gdal.GetDriverByName("GTiff")
    dst_datatype = gdal.GDT_Int16  # gdal.GDT_Float32   #<--- this is fine as we don't actually store elevation data.
    tmp_raster = base_dir + 'mask_' + base_name + '.tif'
    dst_ds = driver.Create(tmp_raster, y, x, 1, dst_datatype)
    dst_ds.SetGeoTransform(src_ds.GetGeoTransform())
    dst_ds.SetProjection(src_ds.GetProjection())
    dst_ds.GetRasterBand(1).SetNoDataValue(dem.GetNoDataValue())
    dst_ds.GetRasterBand(1).WriteArray(Z)
    dst_ds.FlushCache()  # super key to get this to disk
    # noinspection PyUnusedLocal
    dst_ds = None  # close file

    # raster -> polygon
    # subprocess.check_call(
    #     ['%sgdal_polygonize.py %s -b 1 -mask %s -f "ESRI Shapefile" %s' % (gdal_prefix, tmp_raster, tmp_raster,
    #                                                                        base_dir +
    #                                                                        plgs_shp)], shell=True)

    gdal_polygonize(tmp_raster, tmp_raster, base_dir + plgs_shp)

    driver = ogr.GetDriverByName('ESRI Shapefile')
    dataSource = driver.Open(base_dir + plgs_shp, 1)

    # If the input has multiple polygon regions, find the largest polygon and use that as the meshing domain
    layer = dataSource.GetLayer()
    max_geom_area = -1
    max_feature_ID = None
    for feature in layer:
        geom = feature.GetGeometryRef()
        area = geom.GetArea()
        # print 'FID = ' + str(feature.GetFID()) + ' area = ' + str(area)
        if area > max_geom_area:
            max_feature_ID = feature.GetFID()
            max_geom_area = area

    print('Using FID = ' + str(max_feature_ID) + " as the largest continuous area.")
    feats = np.arange(0, layer.GetFeatureCount())
    for f in feats:
        if f != max_feature_ID:
            layer.DeleteFeature(f)
    layer.SyncToDisk()
    dataSource.ExecuteSQL("REPACK " + layer.GetName())
    dataSource.Destroy()

    # allow us to be able to use this directly in the exec_string below if we don't simplify
    outputBufferfn = base_dir + plgs_shp

    # simplify the outter domain constraint and contract by X m, which allows more flexibility in fitting larger
    # triangles to the outter domain.
    if simplify and not no_simplify_buffer:
        # buffering code from http://pcjericks.github.io/py-gdalogr-cookbook/vector_layers.html?highlight=buffer
        inputfn = base_dir + plgs_shp
        outputBufferfn = base_dir + "buffered_" + plgs_shp

        print('Simplifying extents by buffer distance = ' + str(bufferDist))
        inputds = ogr.Open(inputfn)
        inputlyr = inputds.GetLayer()

        shpdriver = ogr.GetDriverByName('ESRI Shapefile')
        if os.path.exists(outputBufferfn):
            shpdriver.DeleteDataSource(outputBufferfn)
        outputBufferds = shpdriver.CreateDataSource(outputBufferfn)
        bufferlyr = outputBufferds.CreateLayer(outputBufferfn, srs_out, geom_type=ogr.wkbPolygon)
        featureDefn = bufferlyr.GetLayerDefn()

        for feature in inputlyr:
            ingeom = feature.GetGeometryRef()
            geomBuffer = ingeom.Buffer(bufferDist)

            outFeature = ogr.Feature(featureDefn)
            outFeature.SetGeometry(geomBuffer)
            bufferlyr.CreateFeature(outFeature)
            outFeature = None
        # close the files
        inputds = None
        outputBufferds = None

    # this needs to be done after we optionally simplify the outer domain. If we do, we need to ensure any interior
    # constraints are clipped to this new domain as the triangulation does not like intersecting constraints.
    for key, data in constraints.items():
        # we need to handle a path being passed in
        output_constraint_fname = os.path.basename(data['file'])
        output_constraint_fname = os.path.splitext(output_constraint_fname)[0]

        df = ogr.Open(data['file'])
        if df is None:
            raise RuntimeError("Unable to open constraint " + data['file'])

        if df.GetLayer(0).GetSpatialRef() is None:
            raise RuntimeError("Constraint " + data['file'] + " must have spatial reference information.")
        df = None

        outname = base_dir + 'constraint_' + output_constraint_fname

        # force all the constraints to have the same extent as the input DEM
        exec_string = '%sogr2ogr -overwrite %s %s  -t_srs \"%s\"' % (
            gdal_prefix, outname + '.shp', data['file'], wkt_out.replace('"', '\\"'))

        if 'simplify' in data:
            exec_string = exec_string + ' -simplify ' + str(
                data['simplify'])  # because of ogr2ogr, the simplification is done in the units of the original data

        subprocess.check_call(exec_string, shell=True)

        # if we simplified with a buffer, clip the constraint shp to that extent. Do this after we project it above.
        if simplify and not no_simplify_buffer:
            clip_outname = base_dir + 'clip_constraint_' + output_constraint_fname
            exec_string = '%sogr2ogr -f "ESRI Shapefile" -clipsrc %s %s %s' % (
                gdal_prefix, outputBufferfn, clip_outname + '.shp', outname + '.shp')  # clip src, output, input
            subprocess.check_call(exec_string, shell=True)

            # update outname to be the clipped one so we can use it below
            outname = clip_outname

        # convert to geoJSON because it's easy to parse
        # ensure it's all line strings and explode the multilines into linestrings
        subprocess.check_call(['%sogr2ogr -f GeoJSON   -nlt LINESTRING -explodecollections  %s %s' % (
            gdal_prefix, outname + '.geojson', outname + '.shp')], shell=True)

        constraints[key]['filename'] = outname + '.geojson'

        # since all the project has already happened in the ogr2ogr step, we can just read it in now
        with open(outname + '.geojson') as f:
            constraints[key]['file'] = json.load(f)

    print('Converting polygon to linestring')
    exec_string = '%sogr2ogr -overwrite %s %s  -nlt LINESTRING' % (
        gdal_prefix, base_dir + 'line_' + plgs_shp, outputBufferfn)

    if simplify:
        exec_string = exec_string + ' -simplify ' + str(simplify_tol)

    subprocess.check_call(exec_string, shell=True)

    # convert to geoJSON because it's easy to parse
    poly_plgs = base_name + '.geojson'
    subprocess.check_call(['%sogr2ogr -f GeoJSON %s %s' % (gdal_prefix, base_dir +
                                                           poly_plgs, base_dir + 'line_' + plgs_shp)], shell=True)

    with open(base_dir + poly_plgs) as f:
        plgs = json.load(f)

    # look through all the features and find the biggest
    idx = -1
    i = 0
    cmax = -1
    l = 0

    for features in plgs['features']:
        if features['geometry'] is None:
            print(
                'Error: the choice of simplify buffer and tolerance has resulted in oversimplifying the domain. '
                'Please choose a tighter tolerance')
            exit(1)

        if features['geometry']['type'] == 'LineString':
            l = len(features['geometry']['coordinates'])
        elif features['geometry']['type'] == 'MultiLineString':
            coords = []

            idx_ml = 0
            len_ml = -1
            j = 0
            # find the largest of the multi lines, ignore holes!
            for lines in features['geometry']['coordinates']:
                l = len(lines)
                if l > len_ml:
                    len_ml = l
                    idx_ml = j
                j += 1

            for l in features['geometry']['coordinates'][idx_ml]:
                coords.append(l)

            features['geometry']['type'] = 'LineString'
            features['geometry']['coordinates'] = coords
            l = len(coords)

        if l > cmax:
            cmax = l
            idx = i
        i += 1

    # assuming just the biggest feature is what we want. Need to add in more to support rivers and lakes
    if plgs['features'][idx]['geometry']['type'] != 'LineString':
        raise RuntimeError('Not linestring')

    coords = plgs['features'][idx]['geometry']['coordinates']

    # We can't just insert this into the global PLGS datastructure as we need that to define the convex hull so merge
    # all the constraints into 1 geojson file so we can just load in the main cpp mesher code to define the interior
    # PLGS TODO: in the future this and the convex hull PLGS should all be in 1 PLGS and the Triangle compatibility
    #  should be removed
    interior_PLGS = {
        "type": "FeatureCollection",
        "name": "interior_PLGS",
        "features": []}

    for key, data in constraints.items():  # over each constraint
        for feat in data['file']['features']:  # over the features present in each constraint
            interior_PLGS['features'].append(feat)

    with open(base_dir + 'interior_PLGS.geojson', 'w') as fp:
        json.dump(interior_PLGS, fp)

    # Create the outer PLGS to constrain the triangulation
    poly_file = 'PLGS' + base_name + '.poly'
    with open(base_dir + poly_file, 'w') as f:
        header = '%d 2 0 0\n' % (len(coords))
        f.write(header)
        vert = 1
        for c in coords:
            f.write('%d %17.11f %17.11f\n' % (vert, c[0], c[1]))
            vert = vert + 1

        f.write('\n')
        header = '%d 0\n' % (len(coords))
        f.write(header)

        for i in range(len(coords)):
            if i + 1 == len(coords):  # last time has to loop back to start
                f.write('%d %d %d\n' % (i + 1, i + 1, 1))
            else:
                f.write('%d %d %d\n' % (i + 1, i + 1, i + 2))

        f.write('0\n')

    # if we aren't reusing the mesh, generate a new one
    if not reuse_mesh:
        execstr = '%s --poly-file %s --tolerance %s --raster %s --area %s --min-area %s --error-metric %s --lloyd %d --interior-plgs-file %s' % \
                  (mesher_path,
                   base_dir + poly_file,
                   max_tolerance,
                   base_dir + base_name + '_projected.tif',
                   max_area,
                   min_area,
                   errormetric,
                   lloyd_itr,
                   base_dir + 'interior_PLGS.geojson'
                   )

        if is_geographic:
            execstr += ' --is-geographic true'

        if use_weights:
            execstr += ' --weight %s' % topo_weight
            execstr += ' --weight-threshold %s' % weight_threshold

        for key, data in parameter_files.items():
            if 'tolerance' in data:
                if data['method'] == 'mode':
                    execstr += ' --category-raster %s --category-frac %s' % (data['filename'][0], data[
                        'tolerance'])  # we need [0] on raster as it's been made iterable by this point
                else:
                    execstr += ' --raster %s --tolerance %s' % (data['filename'][0], data['tolerance'])
            if use_weights and 'weight' in data:
                execstr += ' --weight %s' % data['weight']

        for key, data in initial_conditions.items():
            if 'tolerance' in data:
                if data['method'] == 'mode':
                    execstr += ' --category-raster %s --category-frac %s' % (data['filename'], data['tolerance'])
                else:
                    execstr += ' --raster %s --tolerance %s' % (data['filename'][0], data['tolerance'])
            if use_weights and 'weight' in data:
                execstr += ' --weight %s' % data['weight']

        print(execstr)
        subprocess.check_call(execstr, shell=True)

    # some paramters we want to use to constrain the mesh but don't actually want in the output. This let's use
    # remove them
    keys_to_drop = []
    for key, data in parameter_files.items():
        if 'drop' in data and data['drop'] == True:
            keys_to_drop.append(key)
    for key in keys_to_drop:
        print('Dropping parameter ' + key + ' and not writing to mesh due to user request')
        del parameter_files[key]

    # holds our main mesh structure which we will write out to json to read into CHM
    mesh = {'mesh': {}}
    mesh['mesh']['vertex'] = []

    read_header = False

    invalid_nodes = []  # any nodes that are outside of the domain.
    print('Reading nodes')
    with open(base_dir + 'PLGS' + base_name + '.1.node') as f:
        for line in f:
            if '#' not in line:
                if not read_header:
                    header = re.findall(r"[+-]?\d+(?:\.\d+)?", line)
                    num_nodes = int(header[0])
                    read_header = True
                    mesh['mesh']['nvertex'] = num_nodes
                else:
                    items = re.findall(r"[+-]?\d+(?:\.\d+)?", line)
                    mx = float(items[1])
                    my = float(items[2])

                    mz = extract_point(src_ds, mx, my)

                    if mz == dem.GetNoDataValue() or mz is None:
                        invalid_nodes.append(int(items[0]) - 1)

                    mesh['mesh']['vertex'].append([mx, my, mz])

    print('Length of invalid nodes = ' + str(len(invalid_nodes)))

    # read in the neighbour file, triangle topology
    print('Reading in neighbour file')
    read_header = False
    mesh['mesh']['neigh'] = []
    with open(base_dir + 'PLGS' + base_name + '.1.neigh') as elem:
        for line in elem:
            if '#' not in line:
                if not read_header:
                    read_header = True  # skip header, nothing to do here
                else:
                    items = re.findall(r"[+-]?\d+(?:\.\d+)?", line)
                    v0 = int(items[1]) - 1  # convert to zero indexing
                    v1 = int(items[2]) - 1
                    v2 = int(items[3]) - 1

                    mesh['mesh']['neigh'].append([v0, v1, v2])

    read_header = False

    print('Computing parameters and initial conditions')

    mesh['mesh']['elem'] = []
    mesh['mesh']['is_geographic'] = is_geographic

    # need to save the UTM coordinates so-as to be able to generate lat/long of points if needed later (e.g., CHM)
    if not is_geographic:
        mesh['mesh']['proj4'] = srs.ExportToProj4()
        mesh['mesh']['UTM_zone'] = srs.GetUTMZone()  # negative in southern hemisphere

    # holds parameters and initial conditions for CHM
    params = {}
    ics = {}
    for key, data in parameter_files.items():
        params[key] = []

    params['area'] = []
    params['id'] = []

    for key, data in initial_conditions.items():
        ics[key] = []

    # loop through all the triangles and assign the parameter and ic values to the triangle
    start_time = time.perf_counter()
    with open(base_dir + 'PLGS' + base_name + '.1.ele') as elem:
        for line in elem:
            if '#' not in line:
                if not read_header:
                    header = re.findall(r"[+-]?\d+(?:\.\d+)?", line)
                    nelem = int(header[0])
                    mesh['mesh']['nelem'] = nelem
                    read_header = True  # skip header
                else:
                    items = re.findall(r"[+-]?\d+(?:\.\d+)?", line)
                    v0 = int(items[1]) - 1  # convert to zero indexing
                    v1 = int(items[2]) - 1
                    v2 = int(items[3]) - 1

                    # if the node we have is invalid (out side of domain) we can try to fix it by interpolating from
                    # the surrounding nodes. this can happen when the outter domain constraint is effectively the
                    # limit of the DEM and the node ends up *just* outside of the domain due to numerical
                    # imprecision. estimate an invalid node's z coord from this triangles other nodes' z value
                    if v0 in invalid_nodes:
                        z_v1 = mesh['mesh']['vertex'][v1][2]
                        z_v2 = mesh['mesh']['vertex'][v2][2]
                        tmp = [x for x in [z_v1, z_v2] if x != dem.GetNoDataValue()]
                        # print 'found v0'
                        if len(tmp) != 0:
                            mesh['mesh']['vertex'][v0][2] = float(np.mean(tmp))
                            if verbose:
                                print('replaced invalid with ' + str(mesh['mesh']['vertex'][v0]))
                            invalid_nodes = [x for x in invalid_nodes if x != v0]  # remove from out invalid nodes list.

                    if v1 in invalid_nodes:
                        z_v0 = mesh['mesh']['vertex'][v0][2]
                        z_v2 = mesh['mesh']['vertex'][v2][2]
                        tmp = [x for x in [z_v0, z_v2] if x != dem.GetNoDataValue()]
                        # print 'found v1'
                        if len(tmp) != 0:
                            mesh['mesh']['vertex'][v1][2] = float(np.mean(tmp))
                            if verbose:
                                print('replaced invalid with ' + str(mesh['mesh']['vertex'][v1]))
                            invalid_nodes = [x for x in invalid_nodes if x != v1]  # remove from out invalid nodes list.

                    if v2 in invalid_nodes:
                        # print 'found v2'
                        z_v1 = mesh['mesh']['vertex'][v1][2]
                        z_v0 = mesh['mesh']['vertex'][v0][2]
                        tmp = [x for x in [z_v1, z_v0] if x != dem.GetNoDataValue()]
                        if len(tmp) != 0:
                            mesh['mesh']['vertex'][v2][2] = float(np.mean(tmp))
                            if verbose:
                                print('replaced invalid with ' + str(mesh['mesh']['vertex'][v2]))
                            invalid_nodes = [x for x in invalid_nodes if x != v2]  # remove from out invalid nodes list.
                    mesh['mesh']['elem'].append([v0, v1, v2])

    if len(invalid_nodes) > 0:
        errstr = 'Length of invalid nodes after correction= ' + str(len(invalid_nodes))
        errstr += 'This will have occurred if an entire triangle is outside of the domain. There is no way to ' \
                  'reconstruct this triangle. '
        errstr += 'Try reducing simplify_tol.'
        raise RuntimeError(errstr)

    gt = src_ds.GetGeoTransform()

    tris = range(mesh['mesh']['nelem'])

    ret_tri = []
    ret_tri_ic = []

    csize = len(tris) // nworkers
    if csize < 1:
        csize = 1

    with futures.ProcessPoolExecutor(max_workers=nworkers) as executor:
        for p, i in executor.map(
                partial(do_parameterize, gt, is_geographic, mesh, parameter_files, initial_conditions,
                        src_ds.RasterXSize, src_ds.RasterYSize, srs_out.ExportToProj4()), tris,
                chunksize=csize):
            ret_tri.append(p)
            ret_tri_ic.append(i)

    for key, data in parameter_files.items():
        parameter_files[key]['file'] = None

    for key, data in initial_conditions.items():
        initial_conditions[key]['file'] = None

    for t in ret_tri:
        for key, data in t.items():
            params[key].append(data)

    for t in ret_tri_ic:
        for key, data in t.items():
            ics[key].append(data)

    print('Total time took %s s' % str(round(time.perf_counter() - start_time, 2)))

    if output_write_vtu:
        print('Writing vtu file...')
        write_vtu(base_dir + base_name + '.vtu', mesh, params, ics)

    if output_write_shp:
        print('Writing shp file...')
        write_shp(base_dir + base_name + '_USM.shp', mesh, params, ics)

    print('Saving mesh to file ' + base_name + '.mesh')
    with open(user_output_dir + base_name + '.mesh', 'w') as outfile:
        json.dump(mesh, outfile, indent=4)

    print('Saving parameters to file ' + base_name + '.param')
    with open(user_output_dir + base_name + '.param', 'w') as outfile:
        json.dump(params, outfile, indent=4)

    print('Saving initial conditions  to file ' + base_name + '.ic')
    with open(user_output_dir + base_name + '.ic', 'w') as outfile:
        json.dump(ics, outfile, indent=4)
    print('Done')


def do_parameterize(gt, is_geographic, mesh, parameter_files, initial_conditions, RasterXSize, RasterYSize, srs_proj4,
                    elem):
    params = {}
    ics = {}
    srs_out = osr.SpatialReference()

    srs_out.ImportFromProj4(srs_proj4)

    for key, data in parameter_files.items():
        if data['file'] is None:
            parameter_files[key]['file'] = []
            for f in data['filename']:
                ds = gdal.Open(f)
                if ds is None:
                    raise RuntimeError('Error: Unable to open raster for: %s' % key)
                parameter_files[key]['file'].append(ds)

    for key, data in initial_conditions.items():
        if data['file'] is None:
            initial_conditions[key]['file'] = []
            for f in data['filename']:
                ds = gdal.Open(f)
                if ds is None:
                    raise RuntimeError('Error: Unable to open raster for: %s' % key)
                initial_conditions[key]['file'].append(ds)

    params['id'] = elem

    v0 = mesh['mesh']['elem'][elem][0]
    v1 = mesh['mesh']['elem'][elem][1]
    v2 = mesh['mesh']['elem'][elem][2]

    # Create a temporary vector layer in memory
    mem_drv = ogr.GetDriverByName('Memory')
    mem_ds = mem_drv.CreateDataSource('out')
    mem_layer = mem_ds.CreateLayer('poly', srs_out, ogr.wkbPolygon)

    # we need this to do the area calculation
    ring = ogr.Geometry(ogr.wkbLinearRing)
    ring.AddPoint(mesh['mesh']['vertex'][v0][0], mesh['mesh']['vertex'][v0][1])
    ring.AddPoint(mesh['mesh']['vertex'][v1][0], mesh['mesh']['vertex'][v1][1])
    ring.AddPoint(mesh['mesh']['vertex'][v2][0], mesh['mesh']['vertex'][v2][1])
    ring.AddPoint(mesh['mesh']['vertex'][v0][0],
                  mesh['mesh']['vertex'][v0][1])  # add again to complete the ring.

    # need this for the area calculation
    tpoly = ogr.Geometry(ogr.wkbPolygon)
    tpoly.AddGeometry(ring)

    feature = ogr.Feature(mem_layer.GetLayerDefn())
    feature.SetGeometry(tpoly)

    mem_layer.CreateFeature(feature)

    area = 0
    # if the output is geographic, we need to project to get a reasonable area
    if is_geographic:
        #use an equal area mollweide projection
        srs_moll = osr.SpatialReference()

        srs_moll.ImportFromProj4("+proj=moll +lon_0=0 +x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +units=m +no_defs ")

        # go from what we are outputting (which is geographic) to moll to compute the area
        transform = osr.CoordinateTransformation(srs_out, srs_moll)
        p = tpoly.Clone()
        p.Transform(transform)

        area = p.GetArea()
    else:
        area = tpoly.GetArea()

    params['area'] = area

    # calculate new geotransform of the feature subset
    geom = feature.geometry()
    src_offset = bbox_to_pixel_offsets(gt, geom.GetEnvelope(), RasterXSize, RasterYSize)
    new_gt = (
        (gt[0] + (src_offset[0] * gt[1])),
        gt[1],
        0.0,
        (gt[3] + (src_offset[1] * gt[5])),
        0.0,
        gt[5]
    )

    # get the value under each triangle from each parameter file
    for key, data in parameter_files.items():
        output = []

        for f, m in zip(data['file'], data['method']):
            output.append(rasterize_elem(f, mem_layer, m, srs_out, new_gt, src_offset))

        if 'classifier' in data:
            fn = cloudpickle.loads(data['classifier'])
            output = fn(*output)
        else:
            output = output[0]  # flatten the list for the append below

        # we want to write actual NaN to vtu for better displaying
        if output == -9999:
            output = float('nan')

        if output is None and 'classifier' in data:
            print(f'Error: The user-supplied classifier function for {key} returned None which is not valid.')
            exit(1)

        params[key] = output

    for key, data in initial_conditions.items():
        output = []

        for f, m in zip(data['file'], data['method']):
            output.append(rasterize_elem(f, mem_layer, m, srs_out, new_gt, src_offset))

        if 'classifier' in data:
            output = cloudpickle.loads(data['classifier'])(*output)
        else:
            output = output[0]  # flatten the list for the append below

        if output == -9999:
            output = float('nan')

        if output is None and 'classifier' in data:
            print(f'Error: The user-supplied classifier function for {key} returned None which is not valid.')
            exit(1)

        ics[key] = output

    return params, ics


def regularize_inputs(base_dir, exec_str, gdal_prefix, input_files, pixel_height, pixel_width, srs_out,
                      xmax, xmin, ymax, ymin, fill_holes, max_workers):
    # ensure all the weights sum to 1
    total_weights = 0
    use_weights = False
    param_args = []
    ret = []
    for key, data in input_files.items():
        if 'weight' in data:
            total_weights += data['weight']
            use_weights = True
        if 'classifier' in data:
            input_files[key]['classifier'] = cloudpickle.dumps(input_files[key]['classifier'])

        param_args.append((base_dir, data, exec_str, gdal_prefix, key, pixel_height,
                           pixel_width, srs_out.ExportToProj4(), xmax, xmin, ymax, ymin, fill_holes))

    with futures.ProcessPoolExecutor(max_workers=max_workers) as executor:
        for (r) in executor.map(_future_regularize_inputs, param_args):
            ret.append(r)
    for r in ret:
        key = r['key']
        input_files[key]['filename'] = r['filename']
        input_files[key]['file'] = None

        if not isinstance(input_files[key]['method'], list):
            input_files[key]['method'] = [input_files[key]['method']]

    return total_weights, use_weights


def _future_regularize_inputs(args):
    base_dir, data, exec_str, gdal_prefix, key, pixel_height, pixel_width, srs_out, xmax, xmin, ymax, ymin, fill_holes = args

    # make a copy as this exec string is reused for inital conditions
    # and we don't want the changes made here to impact it
    estr = exec_str
    if data['method'] == 'mode':
        estr = exec_str + 'mode'
    else:
        estr = exec_str + 'average'

    do_cell_resize = True
    estr = estr + ' -tr %s %s'
    # there can be multiple files per param output that we use a classifier to merge into one.
    # we need to process each one. Also you can't use a tolerance with the merging classifier as that makes no sense
    if isinstance(data['file'], list) and len(data['file']) == 1:
        print('Error @ ' + key + ': Do not use [ ] for a single file.')
        exit(-1)
    if not 'classifier' in data and isinstance(data['file'], list):
        print(
            'Error @ ' + key + ': If multiple input files are specified for a single parameter a classifer must be provided.')
        exit(-1)
    if 'tolerance' in data and isinstance(data['file'], list):
        print(
            'Error @ ' + key + ': If multiple input files are specified for a single parameter this cannot be used with a tolerance.')
        exit(-1)
    if isinstance(data['file'], list) and not isinstance(data['method'], list):
        print(
            'Error @ ' + key + ': If multiple input files are specified for a single parameter you need to specify each aggregation method.')
        exit(-1)
    if not isinstance(data['file'], list) and isinstance(data['method'], list):
        print(
            'Error @ ' + key + ': If a single input file is specified you cannot specify multiple aggregation methods.')
        exit(-1)
    if not isinstance(data['file'], list):
        data['file'] = [data['file']]
    if not isinstance(data['method'], list):
        data['method'] = [data['method']]

    ret_df = dict()
    ret_df['filename'] = []
    ret_df['key'] = key

    for f in data['file']:
        # we need to handle a path being passed in
        output_param_fname = os.path.basename(f)
        output_param_fname = os.path.splitext(output_param_fname)[0]

        if gdal.Open(f).GetProjection() == '':
            print("Parameter " + f + " must have spatial reference information.")
            exit(1)

        mangle = uuid.uuid4().hex[:8]

        out_name = base_dir + output_param_fname + '_' + mangle + '_projected.tif'

        # force all the parameter files to have the same extent as the input DEM
        subprocess.check_call([estr % (gdal_prefix,
                                       f, out_name, srs_out, xmin, ymin, xmax, ymax, pixel_width,
                                       pixel_height)], shell=True)
        if fill_holes:
            dist = max([pixel_height,pixel_width]) * 5
            exec_str = '%sgdal_fillnodata.py %s -md %d' % (gdal_prefix, out_name, dist)
            subprocess.check_call([exec_str], shell=True)

        ret_df['filename'].append(out_name)

    return ret_df


# Print iterations progress
# http://stackoverflow.com/a/34325723/410074
def printProgress(iteration, total, prefix='', suffix='', decimals=2, barLength=100):
    """
    Call in a loop to create terminal progress bar
    @params:
        iterations  - Required  : current iteration (Int)
        total       - Required  : total iterations (Int)
        prefix      - Optional  : prefix string (Str)
        suffix      - Optional  : suffix string (Str)
    """
    filledLength = int(round(barLength * iteration / float(total)))
    percents = round(100.00 * (iteration / float(total)), decimals)
    bar = '#' * filledLength + '-' * (barLength - filledLength)
    sys.stdout.write('%s [%s] %s%s %s\r' % (prefix, bar, percents, '%', suffix)),
    sys.stdout.flush()
    if iteration == total:
        print("\n")


# Zonal stats from here https://gist.github.com/perrygeo/5667173
def bbox_to_pixel_offsets(gt, bbox, rasterXsize, rasterYsize):
    originX = gt[0]
    originY = gt[3]
    pixel_width = gt[1]
    pixel_height = gt[5]
    x1 = int((bbox[0] - originX) / pixel_width)
    x2 = int((bbox[1] - originX) / pixel_width) + 1

    y1 = int((bbox[3] - originY) / pixel_height)
    y2 = int((bbox[2] - originY) / pixel_height) + 1

    xsize = x2 - x1
    ysize = y2 - y1

    # only apply this correction if we are touching the underlying raster.
    if x1 < rasterXsize and y1 < rasterYsize:
        # deal with small out of bounds
        if x1 < 0:
            x1 = 0

        if y1 < 0:
            y1 = 0

        if x1 + xsize > rasterXsize:
            xsize = rasterXsize - x1

        if y1 + ysize > rasterYsize:
            ysize = rasterYsize - y1

    return x1, y1, xsize, ysize


def extract_point(raster, mx, my):
    # Convert from map to pixel coordinates.
    # Only works for geotransforms with no rotation.

    rb = raster.GetRasterBand(1)
    gt = raster.GetGeoTransform()

    px = int((mx - gt[0]) / gt[1])  # x pixel
    py = int((my - gt[3]) / gt[5])  # y pixel

    # boundary verticies from Triangle can end up outside of the domain by 1 pixel.
    # if we adjusted back by  1 pixel to get the dz value, it's no problem and still gives a good boundary
    if px == raster.RasterXSize:
        px = px - 1
    if py == raster.RasterYSize:
        py = py - 1
    mz = rb.ReadAsArray(px, py, 1, 1)
    if mz is None:
        return rb.GetNoDataValue()
    mz = float(mz.flatten()[0])

    if mz == rb.GetNoDataValue():

        # look at the surrounding 8 cells
        #  (px-1, py - 1)  (px, py - 1)  (px + 1,py - 1)
        #  (px-1, py    )    (px, py)      (px + 1, py   )
        #  (px-1, py + 1)  (px, py + 1)  (px + 1, py + 1)

        coords = [
            (px - 1, py - 1), (px, py - 1), (px + 1, py - 1),
            (px - 1, py),     (px, py    ), (px + 1, py    ),
            (px - 1, py + 1), (px, py + 1), (px + 1, py + 1)
        ]

        zs = []
        for c in coords:
            try:
                z1 = rb.ReadAsArray(c[0], c[1], 1, 1)
                zs.append(z1)
            except RuntimeError as e:
                pass  # out of bounds, pass

        z = [x for x in zs if x != rb.GetNoDataValue() and x is not None ]

        if len(z) == 0:
            # print 'Warning: The point (%s,%s) and its 8-neighbours lies outside of the DEM domain' % (mx, my)
            return rb.GetNoDataValue()

        mz = float(np.mean(z))

    return mz


def rasterize_elem(rds, mem_layer, aggMethod, srs, new_gt, src_offset):
    raster = rds.GetRasterBand(1)
    src_array = raster.ReadAsArray(*src_offset)

    # Rasterize it
    driver = gdal.GetDriverByName('MEM')
    mask = driver.Create('', src_offset[2], src_offset[3], 1, gdal.GDT_Byte)
    mask.SetGeoTransform(new_gt)
    mask.SetProjection(srs.ExportToWkt())
    err = gdal.RasterizeLayer(mask, [1], mem_layer, burn_values=[1], options=['ALL_TOUCHED=TRUE'])
    if err != 0:
        raise Exception("Error rasterizing layer: %s" % err)

    # holds a mask of where the triangle is on the raster
    mask_arr = mask.ReadAsArray()

    # Mask the source data array with our current feature
    src_array[(mask_arr == 0) | (mask_arr == raster.GetNoDataValue())] = np.nan

    output = -9999.0

    if callable(aggMethod):
        output = float(aggMethod(src_array))
        if np.isnan(output):
            output = raster.GetNoDataValue()
    else:
        if aggMethod == 'mode':
            vals, counts = np.unique(src_array, return_counts=True)
            output = float(vals[np.argmax(counts)])

        elif aggMethod == 'mean':
            output = float(np.nanmean(src_array))
        elif aggMethod == 'max':
            output = float(np.nanmax(src_array))
        elif aggMethod == 'min':
            output = float(np.nanmin(src_array))
        else:
            print('\n\nError: unknown data aggregation method %s\n\n' % aggMethod)
            exit(-1)

    # feature.SetField(key, output)

    # testing code
    # if output < 0:
    #     print "Found < 0"
    #
    #     # testing code
    #     tri_id = str(feature.GetField('triangle'))
    #     print "Tri ID = " + tri_id
    #
    #     print masked
    #
    #     mem_drv = ogr.GetDriverByName('ESRI Shapefile')
    #     mem_ds = mem_drv.CreateDataSource(tri_id+'.shp')
    #     mem_layer = mem_ds.CreateLayer('poly', srs, ogr.wkbPolygon)
    #     mem_layer.CreateFeature(feature.Clone())
    #
    #     driver = gdal.GetDriverByName('GTiff')
    #     rvds = driver.Create(tri_id+'.tiff', src_offset[2], src_offset[3], 1, gdal.GDT_Float32)
    #     rvds.SetGeoTransform(new_gt)
    #     rvds.SetProjection(wkt)
    #     outband = rvds.GetRasterBand(1)
    #     outband.WriteArray(src_array)
    #     # gdal.RasterizeLayer(rvds, [1], mem_layer, burn_values=[1], options=['ALL_TOUCHED=TRUE'])
    #
    #     rvds = None
    #     mem_layer = None
    #     exit(1)

    return output


# To remove the dependency on gdal python bindings and thus the scripts, this is a slimmed down version of the
# gdal_polygonize.py code from
# https://github.com/OSGeo/gdal/blob/release/2.4/gdal/swig/python/scripts/gdal_polygonize.py
def gdal_polygonize(src_filename, mask, dst_filename):
    options = []
    src_band_n = 1
    src_ds = gdal.Open(src_filename)

    srcband = src_ds.GetRasterBand(src_band_n)

    mask_ds = gdal.Open(mask)
    maskband = mask_ds.GetRasterBand(1)

    dst_ds = ogr.Open(dst_filename, update=1)
    drv = ogr.GetDriverByName("ESRI Shapefile")
    dst_ds = drv.CreateDataSource(dst_filename)

    srs = None
    if src_ds.GetProjectionRef() != '':
        srs = osr.SpatialReference()

        srs.ImportFromWkt(src_ds.GetProjectionRef())

    dst_layer = dst_ds.CreateLayer('out', geom_type=ogr.wkbPolygon, srs=srs)

    fd = ogr.FieldDefn('DN', ogr.OFTInteger)
    dst_layer.CreateField(fd)
    dst_field = 0

    result = gdal.Polygonize(srcband, maskband, dst_layer, dst_field, options)


def write_shp(fname, mesh, parameter_files, initial_conditions):
    # Create the shape file to hold the triangulation

    driver = ogr.GetDriverByName("ESRI Shapefile")

    try:
        os.remove(fname)  # remove if existing
    except OSError:
        pass

    output_usm = driver.CreateDataSource(fname)

    srs_out = osr.SpatialReference()

    srs_out.ImportFromProj4(mesh['mesh']['proj4'])

    layer = output_usm.CreateLayer('mesh', srs_out, ogr.wkbPolygon)

    for key, value in parameter_files.items():
        layer.CreateField(ogr.FieldDefn(key, ogr.OFTReal))

    for key, value in initial_conditions.items():
        layer.CreateField(ogr.FieldDefn(key, ogr.OFTReal))

    for elem in range(mesh['mesh']['nelem']):
        v0 = mesh['mesh']['elem'][elem][0]
        v1 = mesh['mesh']['elem'][elem][1]
        v2 = mesh['mesh']['elem'][elem][2]

        # we need this to do the area calculation
        ring = ogr.Geometry(ogr.wkbLinearRing)
        ring.AddPoint(mesh['mesh']['vertex'][v0][0], mesh['mesh']['vertex'][v0][1])
        ring.AddPoint(mesh['mesh']['vertex'][v1][0], mesh['mesh']['vertex'][v1][1])
        ring.AddPoint(mesh['mesh']['vertex'][v2][0], mesh['mesh']['vertex'][v2][1])
        ring.AddPoint(mesh['mesh']['vertex'][v0][0],
                      mesh['mesh']['vertex'][v0][1])  # add again to complete the ring.

        # need this for the area calculation
        tpoly = ogr.Geometry(ogr.wkbPolygon)
        tpoly.AddGeometry(ring)

        feature = ogr.Feature(layer.GetLayerDefn())
        feature.SetGeometry(tpoly)

        # for this section,
        # key[0:10] -> if the name is longer, it'll have been truncated when we made the field
        for key, data in parameter_files.items():
            output = data[elem]
            feature.SetField(key[0:10], output)

        for key, data in initial_conditions.items():
            output = data[elem]
            feature.SetField(key[0:10], output)

        layer.CreateFeature(feature)
    output_usm.FlushCache()
    output_usm = None  # close file


def write_vtu(fname, mesh, parameter_files, initial_conditions):
    vtu = vtk.vtkUnstructuredGrid()

    output_vtk = fname
    vtuwriter = vtk.vtkXMLUnstructuredGridWriter()
    vtuwriter.SetFileName(output_vtk)

    # check what version of vtk we are using so we can avoid the api conflict
    # http://www.vtk.org/Wiki/VTK/VTK_6_Migration/Replacement_of_SetInput#Replacement_of_SetInput.28.29_with_SetInputData.28.29_and_SetInputConnection.28.29
    if vtk.vtkVersion.GetVTKMajorVersion() > 5:
        vtuwriter.SetInputData(vtu)
    else:
        vtuwriter.SetInput(vtu)

    vtu_points = vtk.vtkPoints()
    vtu_triangles = vtk.vtkCellArray()

    vtu_points.SetNumberOfPoints(len(mesh['mesh']['vertex']))

    vtu_cells = {'elevation': vtk.vtkFloatArray(),
                 'cellid': vtk.vtkFloatArray(),
                 'area': vtk.vtkFloatArray()
                 }
    vtu_cells['elevation'].SetName('elevation')
    vtu_cells['area'].SetName('area')
    vtu_cells['cellid'].SetName('cellid')

    for key, data in parameter_files.items():
        k = '[param] ' + key
        vtu_cells[k] = vtk.vtkFloatArray()
        vtu_cells[k].SetName(k)

    for key, data in initial_conditions.items():
        k = '[ic] ' + key
        vtu_cells[k] = vtk.vtkFloatArray()
        vtu_cells[k].SetName(k)

    for elem in range(mesh['mesh']['nelem']):
        v0 = mesh['mesh']['elem'][elem][0]
        v1 = mesh['mesh']['elem'][elem][1]
        v2 = mesh['mesh']['elem'][elem][2]

        vtu_points.SetPoint(v0, mesh['mesh']['vertex'][v0][0],
                            mesh['mesh']['vertex'][v0][1],
                            mesh['mesh']['vertex'][v0][2])

        vtu_points.SetPoint(v1, mesh['mesh']['vertex'][v1][0],
                            mesh['mesh']['vertex'][v1][1],
                            mesh['mesh']['vertex'][v1][2])

        vtu_points.SetPoint(v2, mesh['mesh']['vertex'][v2][0],
                            mesh['mesh']['vertex'][v2][1],
                            mesh['mesh']['vertex'][v2][2])

        triangle = vtk.vtkTriangle()
        triangle.GetPointIds().SetId(0, v0)
        triangle.GetPointIds().SetId(1, v1)
        triangle.GetPointIds().SetId(2, v2)

        vtu_triangles.InsertNextCell(triangle)
        vtu_cells['elevation'].InsertNextTuple1((mesh['mesh']['vertex'][v0][2] +
                                                 mesh['mesh']['vertex'][v1][2] +
                                                 mesh['mesh']['vertex'][v2][2]) / 3.)

        vtu_cells['cellid'].InsertNextTuple1(elem)

        area = 1
        vtu_cells['area'].InsertNextTuple1(area)

        for key, data in parameter_files.items():
            output = data[elem]
            vtu_cells['[param] ' + key].InsertNextTuple1(output)

        for key, data in initial_conditions.items():
            output = data[elem]
            vtu_cells['[ic] ' + key].InsertNextTuple1(output)

    vtu.SetPoints(vtu_points)
    vtu.SetCells(vtk.VTK_TRIANGLE, vtu_triangles)
    for p in vtu_cells.values():
        vtu.GetCellData().AddArray(p)

    vtk_proj4 = vtk.vtkStringArray()
    vtk_proj4.SetNumberOfComponents(1)
    vtk_proj4.SetName("proj4")
    vtk_proj4.InsertNextValue(mesh['mesh']['proj4'])

    vtu.GetFieldData().AddArray(vtk_proj4)

    vtuwriter.Write()


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import sys, os
import json
import numpy as np
import scipy.sparse as sparse
import scipy.sparse.csgraph as graph
import scipy.sparse.linalg as linalg

# Set up CL arguments
import argparse
parser = argparse.ArgumentParser()
parser.add_argument("-o", "--outfile", required=False,
                help="File for output (default overwrites input file).")

parser.add_argument("-t", "--type", required=False, choices={"rcm","nd"},
                help="Type of reordering to perform.",
                nargs='?', const="rcm", type=str, default="rcm")

parser.add_argument("-i", "--infile", required=True,
                help="File for input.")

import matplotlib as mpl
mpl.use('AGG')  # non-gui display (much faster)
import matplotlib.pyplot as plt

def append_global_cell_id_to_mesh_file(args):
    """Read dictionary of arguments, find a desired permutation of cell faces, add new permuted ids to json file"""

    with open(args["infile"]) as f:
        mesh = json.load(f)

    if args["type"]=="rcm":
        print(" Performing RCM bandwidth minimization:")
        mesh['mesh']['cell_global_id'] = compute_rcm_permutation(mesh['mesh']['neigh'])
    elif args["type"]=="nd":
        print(" Performing ND fill-in minimization:")
        mesh['mesh']['cell_global_id'] = compute_nd_permutation(mesh['mesh']['neigh'])

    with open(args["outfile"],'w') as f:
        json.dump(mesh, f, indent=4)

    return

############################################################################
# permutation functions
def compute_rcm_permutation(neighbor_list):
    """Computes the permutation that minimizes the bandwidth of the connectivity matrix
    - uses Reverse CutHill-McKee (RCM) algorithm, which needs CSC or CSR sparse matrix format"""

    A = convert_neighbor_list_to_compressed_matrix(neighbor_list, sparse.csr_matrix)

    # Note we are always dealing with undirected (symmetric) graphs
    permutation = graph.reverse_cuthill_mckee(A,symmetric_mode=True)

    print_bandwidth_before_after_permutation(A,permutation)

    return permutation.tolist()

def compute_nd_permutation(neighbor_list):
    """Computes a nested dissection permutation (minimizes fill-in of matrix factors)"""
    A = convert_neighbor_list_to_compressed_matrix(neighbor_list,sparse.csc_matrix)

    # permutation is determined during the factorization
    LUperm = linalg.splu(A)

    # Not really necessary, but curious
    print_bandwidth_before_after_permutation(
        convert_neighbor_list_to_compressed_matrix(neighbor_list, sparse.csr_matrix),
        LUperm.perm_c )

    print("  Non-zeros in factor L: " + str(LUperm.L.nnz))

    return LUperm.perm_c.tolist()

############################################################################

def convert_neighbor_list_to_compressed_matrix(neighbor_list,compressed_format):
    """Get a sparse compressed_format from a list of neighbors (assuming aij input)
    - compressed_format is one of sparse.csr_matrix or sparse.csc_matrix"""

    N = len(neighbor_list)
    i = []
    j = []
    val = []

    # easier to construct in coo format
    count = 0;
    for ii, neigh in enumerate(neighbor_list):
        for jj in neigh:
            if (jj !=-1):
                i.append(ii)
                j.append(jj)
                count=count+1
                val.append( (count % 2) + 1 ) # entries filled with 1s and 2s
    A = sparse.coo_matrix((val,(i,j)),shape=(N,N))

    return compressed_format(A)


## Bandwidth comparison
def print_bandwidth_before_after_permutation(A,permutation):
    """Display bandwidth before and after permutation"""
    orig_band = compute_bandwidth_CSR(A)
    print("  Original bandwidth: ", orig_band)
    plot_mat_connectivity(A,"orig_order.png")
    Aperm = A[permutation,:][:,permutation]

    post_band = compute_bandwidth_CSR(Aperm)
    plot_mat_connectivity(Aperm,"post_order.png")
    print("  Permuted bandwidth: ", post_band)
    return

def compute_bandwidth_CSR(A):
    """Compute the maximum bandwidth from a CSR sparse matrix A"""
    assert(sparse.isspmatrix_csr(A))
    indices = A.indices
    indptr = A.indptr
    bandwidth = 0
    for row in range(A.shape[0]):
        maxind = np.max(indices[indptr[row]:indptr[row+1]])
        if maxind-row > bandwidth:
            bandwidth = maxind-row
    return bandwidth

def plot_mat_connectivity(A,filename,**kwargs):
    """Plot the spy view of the connectivity matrix
    - TODO kwargs can be used to get values to pass into matplotlib functions"""

    fig=plt.figure(figsize=(18,16), dpi= 80)
    plt.spy(A,markersize=0.05)
    Nnz = A.nnz
    Nrow = A.shape[0]
    ratio = Nnz/(Nrow*Nrow)
    plt.title("Nearest neighbour connectivity")
    plt.xlabel("Number of non-zeros: " + str(Nnz) + " (%.3f %%)" %(ratio*100))
    plt.savefig(filename)

if __name__=="__main__":

    # Parse the input arguments
    args = vars(parser.parse_args())
    if (args["outfile"]==None):
        args["outfile"] = args["infile"]

    append_global_cell_id_to_mesh_file(args)



from skbuild import setup
import subprocess
from packaging.version import LegacyVersion
from skbuild.exceptions import SKBuildError
from skbuild.cmaker import get_cmake_version
import os

def get_installed_gdal_version():
    try:
        version = subprocess.run(["gdal-config","--version"], stdout=subprocess.PIPE).stdout.decode()

        version = version.replace('\n', '')
        version = "=="+version+".*"
        return version
    except FileNotFoundError as e:
        raise(""" ERROR: Could not find the system install of GDAL. 
                  Please install it via your package manage of choice.
                """
            )

# Add CMake as a build requirement if cmake is not installed or is too low a version
# https://scikit-build.readthedocs.io/en/latest/usage.html#adding-cmake-as-building-requirement-only-if-not-installed-or-too-low-a-version
setup_requires = []
try:
    if LegacyVersion(get_cmake_version()) < LegacyVersion("3.16"):
        setup_requires.append('cmake')
except SKBuildError:
    setup_requires.append('cmake')


USE_CONAN = False
try:
  USE_CONAN = os.environ["USE_CONAN"]
except KeyError as e:
  pass # it's ok we don't have this

USE_CONAN = str(USE_CONAN).upper() 

setup(name='mesher',
      version='1.5.1',
      description='Landsurface model mesh generation',
      long_description="""
      Mesher is a novel multi-objective unstructured mesh generation software that allows mesh generation to be generated from an arbitrary number of hydrologically important features while maintaining a variable spatial resolution. 
      Triangle quality is guaranteed as well as a smooth graduation from small to large triangles. Including these additional features resulted in a better representation of spatial heterogeneity versus classic topography-only mesh generation.
      The paper describing mesher can be `found here <https://www.usask.ca/hydrology/papers/Marsh,_et_al_2018.pdf>`_

        Key points
        
        *  A novel multi-objective unstructured mesh generation software, Mesher
        *  Heterogeneity in topography is resolved as well as hydrologically important surface and sub-surface attributes
        *  Spatial heterogeneity is better preserved compared to existing mesh generators
      """,
      author='Chris Marsh',
      author_email='chris.marsh@usask.ca',
      url="https://github.com/Chrismarsh/mesher",
      include_package_data=True,
      cmake_args=['-DCMAKE_BUILD_TYPE:STRING=Release',
                  '-DUSE_CONAN:BOOL='+USE_CONAN],
      scripts=["mesher.py","tools/mesh2vtu.py", "tools/meshmerge.py","tools/meshpermutation.py","tools/meshstats.py"],
      install_requires=['vtk','pygdal-chm'+get_installed_gdal_version(),'numpy','scipy','matplotlib','cloudpickle'],
      setup_requires=setup_requires,
      python_requires='>=3.7, <3.9'
     )


from conans import ConanFile, CMake, tools
import os

class CHMConan(ConanFile):
    settings = "os", "compiler", "build_type", "arch"
   

    name = "mesher"
    version = "1.0"
    license = "https://github.com/Chrismarsh/CHM/blob/master/LICENSE"
    author = "Chris Marsh"
    url = "https://github.com/Chrismarsh/CHM"
    description = "Canadian hydrological model"
    generators = "cmake_find_package"

    options = {"verbose_cmake":[True,False], "build_tests":[True,False] }

    default_options = {"gperftools:heapprof":True,
                       "verbose_cmake":False,
                       "build_tests":True}

    def source(self):
        git = tools.Git()
        git.clone("https://github.com/Chrismarsh/mesher.git",branch=branch)


    def requirements(self):
        self.requires( "cgal/[>=5.2]@CHM/stable" )
        self.requires( "boost/[>=1.75.0]@CHM/stable" )
        self.requires( "gdal/[>=3]@CHM/stable" )
        self.requires("proj/[>=7]@CHM/stable")

    def _configure_cmake(self):
        cmake = CMake(self)

        cmake.configure(source_folder=self.source_folder)

        return cmake

    def build(self):
        cmake = self._configure_cmake()
        cmake.build()
        cmake.test(target="check")

    def package(self):
        cmake = self._configure_cmake()
        cmake.install()


    def imports(self):
        self.copy("*.so*", dst="lib", src="lib")  # From bin to bin
        self.copy("*.dylib*", dst="lib", src="lib")  # From lib to bin
